from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union, Dict, Any
import requests
from rich import print
from rich.console import Group
from rich.live import Live
from rich.markup import escape
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)
from geovisio_cli import utils
from geovisio_cli.auth import login
from geovisio_cli.exception import CliException, raise_for_status
from geovisio_cli.model import (
    Geovisio,
    ManySequences,
    Picture,
    SortMethod,
    BROKEN_UPLOAD_STATUS,
    DUPLICATE_STATUS,
)
from geovisio_cli.sequences.process.standard import SplitParams, MergeParams, process
from geovisio_cli.sequences.status import get_status, info, wait_for_sequence


@dataclass
class UploadError:
    position: int
    picture_path: str
    error: Union[str, dict]
    status_code: int


@dataclass
class UploadReport:
    location: str
    uploaded_pictures: List[Picture] = field(default_factory=lambda: [])
    skipped_pictures: List[Picture] = field(default_factory=lambda: [])
    errors: List[UploadError] = field(default_factory=lambda: [])


def _get_overriden_metadata(picture: Picture):
    """
    Convert the overriden metadata into geovisio API parameters
    """
    from datetime import datetime, timezone

    res: Dict[str, Any] = {}
    m = picture.overridden_metadata
    if m is None:
        return res

    if m.lon is not None:
        res["override_longitude"] = m.lon
    if m.lat is not None:
        res["override_latitude"] = m.lat
    if m.ts is not None:
        # date are send as iso 3339 formated datetime (like '2017-07-21T17:32:28Z')
        res["override_capture_time"] = datetime.fromtimestamp(
            m.ts, timezone.utc
        ).isoformat()
    if len(m.exif) > 0:
        for k in m.exif:
            res[f"override_{k}"] = m.exif[k]

    return res


def _publish(
    session: requests.Session,
    sequences: ManySequences,
    sequenceId: int,
    geovisio: Geovisio,
    wait: bool,
    alreadyBlurred: bool,
    pictureUploadTimeout: float,
) -> UploadReport:
    sequence = sequences.sequences[sequenceId]

    # Read sequence data
    if sequence.id:
        sequence = info(session, sequence)
        print(
            f'📡 Resuming upload of sequence "{sequence.title}" (part {sequenceId+1}/{len(sequences.sequences)})'
        )
        print(f"  - Folder: {sequence.path}")
        print(f"  - API ID: {sequence.id}")
    else:
        print(
            f'📡 Uploading sequence "{sequence.title}" (part {sequenceId+1}/{len(sequences.sequences)})'
        )
        print(f"  - Folder: {sequence.path}")

    data = {}
    if sequence.title:
        data["title"] = sequence.title

    # List pictures to upload
    picturesToUpload: Dict[int, Picture] = {}
    picturesSkipped: List[Picture] = []
    for i, p in enumerate(sequence.pictures, start=1):
        if p.status == DUPLICATE_STATUS:
            picturesSkipped.append(p)
        elif p.id is None or p.status == BROKEN_UPLOAD_STATUS:
            picturesToUpload[i] = p

    # Create sequence on initial publishing
    if not sequence.id:
        seq = session.post(
            f"{geovisio.url}/api/collections", data=data, timeout=utils.REQUESTS_TIMEOUT
        )
        raise_for_status(seq, "Impossible to query GeoVisio")

        sequence.id = seq.json()["id"]
        sequence.location = seq.headers["Location"]
        sequences.persist()

        print(f"\n  ✅ Created collection {sequence.location}")

    else:
        if len(picturesToUpload) == 0:
            print(
                f"\n  ✅ Everything ({len(sequence.pictures)} picture{'s' if len(sequence.pictures) != 1 else ''}) have already been uploaded, nothing to do"
            )
            assert sequence.location
            return UploadReport(
                location=sequence.location, skipped_pictures=picturesSkipped
            )
        print(
            f"\n  ⏭️ Skipping {len(sequence.pictures) - len(picturesToUpload)} already published pictures"
        )

    if not sequence.location:
        raise CliException("Sequence has no API location defined")

    report = UploadReport(location=sequence.location, skipped_pictures=picturesSkipped)

    uploading_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(),
        TimeElapsedColumn(),
        TextColumn("[{task.completed}/{task.total}]"),
    )
    current_pic_progress = Progress(
        TextColumn("  📷 Processing [bold purple]{task.fields[file]}"),
        SpinnerColumn("simpleDots"),
    )
    error_progress = Progress(TextColumn("{task.description}"))

    last_error = Progress(
        TextColumn("🔎 Last error 🔎\n{task.description}"),
    )
    error_panel = Panel(Group(error_progress, last_error), title="Errors")
    uploading_task = uploading_progress.add_task(
        f"  [green]🚀 Uploading pictures...",
        total=len(picturesToUpload),
    )
    current_pic_task = current_pic_progress.add_task("", file="")
    progress_group = Group(
        uploading_progress,
        current_pic_progress,
        error_panel,
    )
    error_task = error_progress.add_task("[green]No errors")
    last_error_task = last_error.add_task("", visible=False)
    with Live(progress_group):
        for i, p in picturesToUpload.items():
            if not p.path:
                raise CliException(f"Missing path for picture {str(p)}")

            uploading_progress.advance(uploading_task)
            current_pic_progress.update(current_pic_task, file=p.path.split("/")[-1])

            post_data = {
                "position": i,
                "isBlurred": "true" if alreadyBlurred else "false",
            }
            post_data.update(_get_overriden_metadata(p))
            try:
                picture_response = session.post(
                    f"{sequence.location}/items",
                    files={"picture": open(p.path, "rb")},
                    data=post_data,
                    timeout=(utils.REQUESTS_CNX_TIMEOUT, pictureUploadTimeout),
                )
            except (requests.Timeout,) as timeout_error:
                raise CliException(
                    f"""Timeout while trying to post picture. Maybe the instance is overloaded? Please contact your instance administrator.

            [bold]Error:[/bold]
            {timeout_error}"""
                )
            except (
                requests.ConnectionError,
                requests.ConnectTimeout,
                requests.TooManyRedirects,
            ) as cnx_error:
                raise CliException(
                    f"""Impossible to reach GeoVisio while trying to post a picture, connection was lost. Please contact your instance administrator.

            [bold]Error:[/bold]
            {cnx_error}"""
                )

            # Picture at given position exists -> mark it as OK
            if picture_response.status_code == 409:
                sequence = get_status(session, sequence)
                sequences.persist()
                report.uploaded_pictures.append(p)

            elif picture_response.status_code >= 400:
                body = picture_response.text

                # Format a better text if JSON details are available
                if picture_response.headers.get("Content-Type") == "application/json":
                    bodyJson = picture_response.json()
                    body = bodyJson.get("message")
                    if "details" in bodyJson and "error" in bodyJson["details"]:
                        body += "\n" + bodyJson["details"]["error"]

                report.errors.append(
                    UploadError(
                        position=i,
                        picture_path=p.path,
                        status_code=picture_response.status_code,
                        error=body,
                    )
                )

                error_progress.update(
                    error_task,
                    description=f"[bold red]{len(report.errors)} errors",
                )
                last_error.update(last_error_task, description=body, visible=True)
                p.status = BROKEN_UPLOAD_STATUS
                sequences.persist()

            else:
                p.location = picture_response.headers["Location"]
                p.id = picture_response.json()["id"]
                p.status = None
                report.uploaded_pictures.append(p)
                sequences.persist()

    if not report.uploaded_pictures:
        print(
            f"  [red]💥 All pictures upload of sequence {sequence.title} failed! 💥[/red]"
        )
    else:
        print(
            f"  🎉 [bold green]{len(report.uploaded_pictures)}[/bold green] pictures uploaded"
        )
    if report.errors:
        print(f"  [bold red]{len(report.errors)}[/bold red] pictures not uploaded:")
        for e in report.errors:
            msg: Union[str, dict] = e.error
            if isinstance(e.error, str):
                spacing = "\n     "
                msg = escape(e.error.replace("\n", spacing))
            print(
                f" - {e.picture_path} (HTTP status [bold]{e.status_code}[/]):{spacing}{msg}"
            )

    if wait:
        wait_for_sequence(session, sequence)
    else:
        print(f"  Note: You can follow the picture processing with the command:")

        print(
            f"    [bold]geovisio collection-status --wait --location {sequence.location}"
        )
    return report


def _login_if_needed(session: requests.Session, geovisio: Geovisio) -> bool:
    # Check if API needs login
    apiConf = session.get(f"{geovisio.url}/api/configuration")
    if apiConf.json().get("auth", {}).get("enabled", False):
        logged_in = login(session, geovisio)
        if not logged_in:
            return False
    return True


def upload(
    path: Path,
    geovisio: Geovisio,
    title: Optional[str],
    pictureUploadTimeout: float,
    wait: bool = False,
    alreadyBlurred: bool = False,
    sortMethod: Optional[SortMethod] = None,
    splitParams: Optional[SplitParams] = None,
    mergeParams: Optional[MergeParams] = None,
    disableCertCheck=False,
) -> List[UploadReport]:
    # early test that the given url is correct
    with utils.createSessionWithRetry() as s:
        if disableCertCheck:
            s.verify = False
        utils.test_geovisio_url(s, geovisio.url)
        # early test login
        if not _login_if_needed(s, geovisio):
            raise CliException(
                "🔁 Computer not authenticated yet, impossible to upload pictures, but you can try again the same upload command after finalizing the login"
            )

        ms = process(path, title, sortMethod, splitParams, mergeParams)
        res = []
        for i in range(len(ms.sequences)):
            print("")  # Do not remove, gives a bit of spacing in output
            res.append(
                _publish(
                    session=s,
                    sequences=ms,
                    sequenceId=i,
                    geovisio=geovisio,
                    wait=wait,
                    alreadyBlurred=alreadyBlurred,
                    pictureUploadTimeout=pictureUploadTimeout,
                )
            )

        return res



import pytest
import os
import shutil
from geovisio_cli.sequences.process import standard
from .conftest import FIXTURE_DIR
from pathlib import Path, PurePath
from geovisio_cli.sequences.external_metadata import MetadataHandler
from geovisio_cli.model import DUPLICATE_STATUS
from geopic_tag_reader.reader import PartialGeoPicTags


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
    os.path.join(FIXTURE_DIR, "not_a_pic.md"),
)
def test_upload_with_invalid_file(datafiles):
    ms = standard.process(path=Path(datafiles), title=None)

    assert len(ms.sequences) == 1
    s = ms.sequences[0]
    assert len(s.pictures) == 3
    assert [PurePath(p.path).stem for p in s.pictures] == ["e1", "e2", "e3"]
    assert s.title == Path(datafiles).name


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_load_pictures_without_metadata_handler(datafiles):
    files = [
        datafiles / "e1.jpg",
        datafiles / "e2.jpg",
        datafiles / "e3.jpg",
    ]
    pictures = standard._load_pictures(picture_files=files, metadata_handler=None)

    assert len(pictures) == 3

    assert pictures[0].path == str(datafiles / "e1.jpg")
    assert pictures[0].metadata
    assert pictures[0].metadata.lon == -1.6844680555555556
    assert pictures[0].metadata.lat == 48.15506638888889
    assert pictures[0].metadata.ts == 1666166194.0
    assert pictures[0].metadata.type == "flat"
    assert pictures[0].metadata.make == "SONY"
    assert pictures[0].metadata.model == "FDR-X1000V"

    assert pictures[1].path == str(datafiles / "e2.jpg")
    assert pictures[1].metadata
    assert pictures[1].metadata.lon == -1.684506388888889
    assert pictures[1].metadata.lat == 48.155071388888885
    assert pictures[1].metadata.ts == 1666166196
    assert pictures[1].metadata.type == "flat"
    assert pictures[1].metadata.make == "SONY"
    assert pictures[1].metadata.model == "FDR-X1000V"

    assert pictures[2].path == str(datafiles / "e3.jpg")
    assert pictures[2].metadata
    assert pictures[2].metadata.lon == -1.684546388888889
    assert pictures[2].metadata.lat == 48.155073055555555
    assert pictures[2].metadata.type == "flat"
    assert pictures[2].metadata.model == "FDR-X1000V"
    assert pictures[2].metadata.ts == 1666166198.0
    assert pictures[2].metadata.make == "SONY"


class SimpleMetadataHandler(MetadataHandler):
    """Helper class to emulate metadata handlers"""

    def __init__(self, metadata) -> None:
        super().__init__()
        self.metadata = metadata

    @staticmethod
    def new_from_file(_):
        pass

    def get(self, file_path):
        return self.metadata.get(file_path)


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_load_pictures_with_metadata_handler(datafiles):
    files = [
        datafiles / "e1.jpg",
        datafiles / "e2.jpg",
        datafiles / "e3.jpg",
    ]

    simple_metadata_handler = SimpleMetadataHandler(
        {
            datafiles / "e1.jpg": PartialGeoPicTags(ts=4242, make="CANON"),
            # no metadata for `e2`, should be a problem
            datafiles / "e3.jpg": PartialGeoPicTags(lon=42),
        }
    )

    pictures = standard._load_pictures(
        picture_files=files, metadata_handler=simple_metadata_handler
    )

    assert len(pictures) == 3

    assert pictures[0].path == str(datafiles / "e1.jpg")
    assert pictures[0].status is None  # None means everything is fine for the moment
    assert pictures[0].metadata
    assert pictures[0].metadata.lon == -1.6844680555555556
    assert pictures[0].metadata.lat == 48.15506638888889
    assert pictures[0].metadata.type == "flat"
    assert pictures[0].metadata.model == "FDR-X1000V"
    assert pictures[0].metadata.ts == 1666166194.0
    assert pictures[0].metadata.make == "SONY"
    assert pictures[0].overridden_metadata
    assert pictures[0].overridden_metadata.ts == 4242
    assert pictures[0].overridden_metadata.make == "CANON"

    assert pictures[1].path == str(datafiles / "e2.jpg")
    assert pictures[0].status is None  # None means everything is fine for the moment
    assert pictures[1].metadata
    assert pictures[1].metadata.lon == -1.684506388888889
    assert pictures[1].metadata.lat == 48.155071388888885
    assert pictures[1].metadata.ts == 1666166196
    assert pictures[1].metadata.type == "flat"
    assert pictures[1].metadata.make == "SONY"
    assert pictures[1].metadata.model == "FDR-X1000V"

    assert pictures[2].path == str(datafiles / "e3.jpg")
    assert pictures[0].status is None  # None means everything is fine for the moment
    assert pictures[2].metadata
    assert pictures[2].metadata.lon == -1.684546388888889
    assert pictures[2].metadata.lat == 48.155073055555555
    assert pictures[2].metadata.type == "flat"
    assert pictures[2].metadata.model == "FDR-X1000V"
    assert pictures[2].metadata.ts == 1666166198.0
    assert pictures[2].metadata.make == "SONY"
    assert pictures[2].overridden_metadata
    assert pictures[2].overridden_metadata.lon == 42


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1_without_exif.jpg"),
    os.path.join(FIXTURE_DIR, "e2_without_coord.jpg"),
    os.path.join(FIXTURE_DIR, "e3_without_exif.jpg"),
)
def test_load_pictures_with_metadata_handler_and_partial_exifs(datafiles):
    """
    Loading picture with partial metadata should be no problem if mandatory metadata can be found in external metada
    Here only e3_without_exif.jpg should be in error since it has no exif tags, and only `ts` defined in external metadata
    """
    files = [
        datafiles / "e1_without_exif.jpg",
        datafiles / "e2_without_coord.jpg",
        datafiles / "e3_without_exif.jpg",
    ]

    simple_metadata_handler = SimpleMetadataHandler(
        {
            datafiles
            / "e1_without_exif.jpg": PartialGeoPicTags(
                ts=4242, lon=12, lat=42, make="CANON"
            ),
            datafiles
            / "e2_without_coord.jpg": PartialGeoPicTags(lon=13, lat=43, make="CANON"),
            datafiles / "e3_without_exif.jpg": PartialGeoPicTags(ts=4243),
        }
    )

    pictures = standard._load_pictures(
        picture_files=files, metadata_handler=simple_metadata_handler
    )

    assert len(pictures) == 3

    assert pictures[0].path == str(datafiles / "e1_without_exif.jpg")
    assert pictures[0].metadata is None
    assert pictures[0].status is None  # None means everything is fine for the moment
    assert pictures[0].overridden_metadata
    assert pictures[0].overridden_metadata.ts == 4242
    assert pictures[0].overridden_metadata.lon == 12
    assert pictures[0].overridden_metadata.lat == 42
    assert pictures[0].overridden_metadata.make == "CANON"

    assert pictures[1].path == str(datafiles / "e2_without_coord.jpg")
    assert pictures[1].status is None
    assert pictures[1].metadata is None
    assert pictures[1].overridden_metadata
    # the overridden_metadata also contains exif tags
    assert pictures[1].overridden_metadata.lon == 13
    assert pictures[1].overridden_metadata.lat == 43
    assert pictures[1].overridden_metadata.ts == 1666166196
    assert pictures[1].overridden_metadata.type == "flat"
    assert (
        pictures[1].overridden_metadata.make == "CANON"
    )  # -< both define in exif and external, we use the external value
    assert pictures[1].overridden_metadata.model == "FDR-X1000V"

    assert pictures[2].path == str(datafiles / "e3_without_exif.jpg")
    assert pictures[2].status == "broken-metadata"
    assert pictures[2].overridden_metadata
    assert pictures[2].overridden_metadata.ts == 4243


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_process_with_duplicates(datafiles):
    # Create a duplicate of e1
    shutil.copyfile(datafiles / "e1.jpg", datafiles / "e0.jpg")

    # Process sequence
    sequences = standard.process(
        Path(datafiles),
        "test",
        mergeParams=standard.MergeParams(maxDistance=1, maxRotationAngle=30),
        sortMethod=standard.SortMethod.filename_asc,
    )

    # Check results (second e1 should be marked as duplicate)
    assert len(sequences.sequences) == 1
    s = sequences.sequences[0]
    assert len(s.pictures) == 4

    assert s.pictures[0].status == None
    assert s.pictures[0].path == datafiles / "e0.jpg"
    assert s.pictures[1].status == DUPLICATE_STATUS
    assert s.pictures[1].path == datafiles / "e1.jpg"
    assert s.pictures[2].status == None
    assert s.pictures[2].path == datafiles / "e2.jpg"
    assert s.pictures[3].status == None
    assert s.pictures[3].path == datafiles / "e3.jpg"



import os
import pytest

import geovisio_cli.sequences.status
import geovisio_cli.sequences.upload
from ..conftest import FIXTURE_DIR
from pathlib import Path
import requests
from geovisio_cli import exception, model
from datetime import timedelta
from geopic_tag_reader.reader import PartialGeoPicTags
from datetime import datetime, timezone


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
    os.path.join(FIXTURE_DIR, "metadata_exif.csv"),
)
def test_valid_csv_upload(geovisio_with_token, datafiles):
    # value stored in metadata.csv
    e1_new_lat = 50.5151
    e1_new_lon = 3.265
    e1_new_exif = {
        "Exif.Image.Software": "MS Paint",
        "Exif.Image.Artist": "A 2 years old",
        "Xmp.xmp.Rating": "1",
    }
    e3_new_lat = 50.513433333
    e3_new_lon = 3.265277778
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        pictureUploadTimeout=20,
    )
    assert len(uploadReports) == 1
    assert len(uploadReports[0].uploaded_pictures) == 3
    assert len(uploadReports[0].errors) == 0

    collection = model.Sequence(location=uploadReports[0].location)
    with requests.session() as s:
        geovisio_cli.sequences.status.wait_for_sequence(
            s, collection, timeout=timedelta(minutes=1)
        )
        status = geovisio_cli.sequences.status.get_status(s, collection)
        sequence_info = geovisio_cli.sequences.status.info(s, collection)
    # 3 pictures should have been uploaded
    assert len(status.pictures) == 3

    for i in status.pictures:
        assert i.status == "ready"

    # the collection should also have 3 items
    collection = requests.get(f"{collection.location}/items")
    collection.raise_for_status()

    features = collection.json()["features"]
    assert len(features) == 3

    assert sequence_info.title == "some title"
    assert sequence_info.interior_orientation == [
        model.InteriorOrientation(make="SONY", model="FDR-X1000V", field_of_view=None)
    ]

    # and the send pictures should have the overriden info
    assert features[0]["geometry"]["coordinates"] == [e1_new_lon, e1_new_lat]
    assert features[0]["properties"]["datetime"] == "2018-01-22T02:52:09+00:00"
    assert features[0]["properties"]["exif"]["Exif.Image.Software"] == "MS Paint"
    assert features[0]["properties"]["exif"]["Exif.Image.Artist"] == "A 2 years old"
    assert features[0]["properties"]["exif"]["Xmp.xmp.Rating"] == "1"
    assert features[2]["geometry"]["coordinates"] == [e3_new_lon, e3_new_lat]
    assert features[2]["properties"]["datetime"] == "2018-01-22T02:54:09+00:00"



import typer
from typing_extensions import Annotated
from pathlib import Path
from geovisio_cli import exception, model, auth, utils, __version__
from rich import print
from rich.panel import Panel
from typing import Optional
import os
import geovisio_cli.sequences.process.standard
import requests

import geovisio_cli.sequences.status
import geovisio_cli.sequences.upload


app = typer.Typer()


def version_callback(value: bool):
    if value:
        print(f"GeoVisio command-line client ([blue bold]v{__version__}[/blue bold])")
        utils.check_if_lastest_version()
        raise typer.Exit()


@app.callback(help=f"GeoVisio command-line client (v{__version__})")
def common(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Show GeoVisio command-line client version and exit",
        ),
    ] = None,
):
    pass


@app.command()
def upload(
    path: Path = typer.Argument(..., help="Local path to your sequence folder"),
    api_url: str = typer.Option(..., help="GeoVisio endpoint URL"),
    user: str = typer.Option(
        default=None,
        hidden=True,
        help="""DEPRECATED: GeoVisio user name if the geovisio instance needs it.
If none is provided and the geovisio instance requires it, the username will be asked during run.
""",
        envvar="GEOVISIO_USER",
    ),
    password: str = typer.Option(
        default=None,
        hidden=True,
        help="""DEPRECATED: GeoVisio password if the geovisio instance needs it.
If none is provided and the geovisio instance requires it, the password will be asked during run.
Note: is is advised to wait for prompt without using this variable.
""",
        envvar="GEOVISIO_PASSWORD",
    ),
    wait: bool = typer.Option(default=False, help="Wait for all pictures to be ready"),
    isBlurred: bool = typer.Option(
        False,
        "--is-blurred/--is-not-blurred",
        help="Define if sequence is already blurred or not",
    ),
    title: Optional[str] = typer.Option(
        default=None,
        help="Collection title. If not provided, the title will be the directory name.",
    ),
    token: Optional[str] = typer.Option(
        default=None,
        help="""GeoVisio token if the geovisio instance needs it.

If none is provided and the geovisio instance requires it, the token will be asked during run.
Note: is is advised to wait for prompt without using this variable.
""",
    ),
    sort_method: Optional[model.SortMethod] = typer.Option(
        default="time-asc",
        help="Strategy used for sorting your pictures. Either by filename or EXIF time, in ascending or descending order.",
    ),
    split_distance: Optional[int] = typer.Option(
        default=100,
        help="Maximum distance between two pictures to be considered in the same sequence (in meters).",
    ),
    split_time: Optional[int] = typer.Option(
        default=60,
        help="Maximum time interval between two pictures to be considered in the same sequence (in seconds).",
    ),
    duplicate_distance: Optional[float] = typer.Option(
        default=1,
        help="Maximum distance between two pictures to be considered as duplicates (in meters).",
    ),
    duplicate_rotation: Optional[int] = typer.Option(
        default=30,
        help="Maximum angle of rotation for two too-close-pictures to be considered as duplicates (in degrees).",
    ),
    picture_upload_timeout: float = typer.Option(
        default=60.0,
        help="Timeout time to receive the first byte of the response for each picture upload (in seconds)",
    ),
    disable_cert_check: Annotated[
        bool,
        typer.Option(
            "--disable-cert-check/--enable-cert-check",
            help="Disable SSL certificates checks while uploading. This should not be used, unless if you -really- know what you are doing.",
        ),
    ] = False,
):
    """Processes and sends a given sequence on your GeoVisio API"""

    def cmd():
        if user or password:
            raise exception.CliException(
                "user/password authentication have been deprecated, use a token or `geovisio login` instead"
            )
        geovisio = model.Geovisio(url=api_url, token=token)
        geovisio_cli.sequences.upload.upload(
            path,
            geovisio,
            wait=wait,
            alreadyBlurred=isBlurred,
            title=title,
            sortMethod=sort_method,
            splitParams=geovisio_cli.sequences.process.standard.SplitParams(
                maxDistance=split_distance,
                maxTime=split_time,
            ),
            mergeParams=geovisio_cli.sequences.process.standard.MergeParams(
                maxDistance=duplicate_distance,
                maxRotationAngle=duplicate_rotation,
            ),
            pictureUploadTimeout=picture_upload_timeout,
            disableCertCheck=disable_cert_check,
        )

    _run_command(cmd, "importing collection", path)


@app.command()
def test_process(
    path: Path = typer.Argument(..., help="Local path to your sequence folder"),
    title: Optional[str] = typer.Option(
        default=None,
        help="Collection title. If not provided, the title will be the directory name.",
    ),
    sort_method: Optional[model.SortMethod] = typer.Option(
        default="time-asc",
        help="Strategy used for sorting your pictures. Either by filename or EXIF time, in ascending or descending order.",
    ),
    split_distance: Optional[int] = typer.Option(
        default=100,
        help="Maximum distance between two pictures to be considered in the same sequence (in meters).",
    ),
    split_time: Optional[int] = typer.Option(
        default=60,
        help="Maximum time interval between two pictures to be considered in the same sequence (in seconds).",
    ),
    duplicate_distance: Optional[float] = typer.Option(
        default=1,
        help="Maximum distance between two pictures to be considered as duplicates (in meters).",
    ),
    duplicate_rotation: Optional[int] = typer.Option(
        default=30,
        help="Maximum angle of rotation for two too-close-pictures to be considered as duplicates (in degrees).",
    ),
):
    """(For testing) Generates a TOML file with metadata used for upload"""

    def cmd():
        geovisio_cli.sequences.process.standard.process(
            path,
            title,
            sortMethod=sort_method,
            splitParams=geovisio_cli.sequences.process.standard.SplitParams(
                maxDistance=split_distance,
                maxTime=split_time,
            ),
            mergeParams=geovisio_cli.sequences.process.standard.MergeParams(
                maxDistance=duplicate_distance,
                maxRotationAngle=duplicate_rotation,
            ),
        )
        outputFile = os.path.join(path, model.SEQUENCE_TOML_FILE)
        print(
            "\n✅ [green]Metadata file saved to: [bold]" + outputFile + "[/bold][/green]"
        )

    _run_command(cmd, "processing collection", path)


@app.command()
def collection_status(
    id: Optional[str] = typer.Option(default=None, help="Id of the collection"),
    api_url: Optional[str] = typer.Option(default=None, help="GeoVisio endpoint URL"),
    location: Optional[str] = typer.Option(
        default=None, help="Full url of the collection"
    ),
    wait: bool = typer.Option(default=False, help="wait for all pictures to be ready"),
    disable_cert_check: Annotated[
        bool,
        typer.Option(
            "--disable-cert-check/--enable-cert-check",
            help="Disable SSL certificates checks while uploading. This should not be used, unless if you -really- know what you are doing.",
        ),
    ] = False,
):
    """
    Print the status of a collection.\n
    Either a --location should be provided, with the full location url of the collection
    or only the --id combined with the --api-url
    """

    def cmd():
        if location is None:
            if api_url is None or id is None:
                raise exception.CliException(
                    "The way to identify the collection should be either with --location or with --id combined with --api-url"
                )
            l = f"{api_url}/api/collections/{id}"
        else:
            l = location

        mySequence = model.Sequence(id=id, location=l)
        with utils.createSessionWithRetry() as s:
            if disable_cert_check:
                s.verify = False
            geovisio_cli.sequences.status.display_sequence_status(s, mySequence)

            if wait:
                geovisio_cli.sequences.status.wait_for_sequence(s, mySequence)

    _run_command(cmd, "getting collection status")


@app.command(
    help=f"""
    Authenticate into the given instance, and save credentials in a configuration file.

    This will generate credentials, and ask the user to visit a page to associate those credentials to the user's account.

    The credentials will be stored in {auth.get_config_file_path()}
    """
)
def login(
    api_url: str = typer.Option(..., help="GeoVisio endpoint URL"),
    disable_cert_check: Annotated[
        bool,
        typer.Option(
            "--disable-cert-check/--enable-cert-check",
            help="Disable SSL certificates checks while uploading. This should not be used, unless if you -really- know what you are doing.",
        ),
    ] = False,
):
    return _run_command(
        lambda: auth.create_auth_credentials(
            model.Geovisio(url=api_url), disable_cert_check
        ),
        "authenticating",
    )


def _run_command(command, command_name_for_error, path: Optional[Path] = None):
    try:
        utils.check_if_lastest_version()
        command()
    except exception.CliException as e:
        print(
            Panel(
                f"{e}",
                title=f"[red]Error while {command_name_for_error}"
                + (f" ({str(path.resolve())})" if path is not None else ""),
                border_style="red",
            )
        )
        return 1



from __future__ import annotations  # used to type hint the return of init_from_file
from pathlib import Path
from typing import Optional, List
from abc import ABC, abstractmethod
from geopic_tag_reader import reader


class MetadataHandler(ABC):
    @staticmethod
    @abstractmethod
    def new_from_file(file_name: Path) -> Optional[MetadataHandler]:
        """
        Method to create the handler from a file.
        If the handler cannot be created from this file, returns None
        """
        pass

    @abstractmethod
    def get(self, file_name: Path) -> Optional[reader.PartialGeoPicTags]:
        """
        Method overloaded by the MetadataHandler

        It takes the path of a picture file and return a PartialGeoPicTags if possible
        """
        return None


def find_handler(path) -> Optional[MetadataHandler]:
    """
    Find a MetadataHandler in the directory.

    Each registered MetadataHandler is responsible to know if they can be inited by a given file
    """
    from . import csv_metadata

    handler_list: List[type[MetadataHandler]] = [
        # add all handlers here
        csv_metadata.CsvMetadataHandler,
    ]
    for f in path.iterdir():
        if not f.is_file():
            continue

        for handler_class in handler_list:
            h = handler_class.new_from_file(f)
            if h:
                return h

    return None



import pytest
from geovisio_cli.sequences.external_metadata import utils
import datetime
from geovisio_cli import exception


def test_parse_capture_time():
    r = utils.parse_capture_time("2018-01-22T02:52:09.999346+00:00")
    assert (
        r
        == datetime.datetime(
            year=2018,
            month=1,
            day=22,
            hour=2,
            minute=52,
            second=9,
            microsecond=999346,
            tzinfo=datetime.timezone.utc,
        ).timestamp()
    )


def test_parse_bad_date_capture_time():
    with pytest.raises(exception.CliException) as e:
        utils.parse_capture_time("plop")
    assert (
        str(e.value)
        == "The capture_time was not recognized (should follow the RFC 3339): plop (Invalid isoformat string: 'plop')"
    )



import pytest
import os
from geovisio_cli import exception, model
import geovisio_cli.sequences.process.standard
import geovisio_cli.sequences.upload
from .conftest import FIXTURE_DIR, MOCK_API_URL
from pathlib import Path
import tomli  # type: ignore
import tomli_w  # type: ignore
import requests
from geopic_tag_reader.reader import PartialGeoPicTags


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "invalid_pic.jpg"),
)
def test_upload_with_no_valid_file(datafiles):
    with pytest.raises(exception.CliException) as e:
        seq = geovisio_cli.sequences.process.standard._read_sequences(Path(datafiles))

    assert e.match("All read pictures have invalid metadata")


def mock_api_post_collection_fail(requests_mock):
    requests_mock.post(
        MOCK_API_URL + "/api/collections",
        exc=requests.exceptions.ConnectTimeout,
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_collection_create_failure(requests_mock, datafiles):
    mock_api_post_collection_fail(requests_mock)

    with pytest.raises(exception.CliException) as e:
        geovisio_cli.sequences.upload.upload(
            path=datafiles,
            geovisio=model.Geovisio(url=MOCK_API_URL),
            title="Test",
            alreadyBlurred=True,
            pictureUploadTimeout=20,
        )

    assert str(e.value).startswith("Error while connecting to the API")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_upload_with_invalid_file(requests_mock, datafiles):
    # Put apart third picture
    os.rename(datafiles + "/e2.jpg", datafiles + "/e2.bak")
    os.rename(datafiles + "/e3.jpg", datafiles + "/e3.bak")

    # Mock collection creation
    gvsMock = model.Geovisio(url=MOCK_API_URL)
    seqId = "123456789"
    picId1 = "123"
    picId2 = "456"
    picId3 = "789"
    requests_mock.get(f"{MOCK_API_URL}/api", json={})
    requests_mock.get(f"{MOCK_API_URL}/api/configuration", json={})
    requests_mock.get(
        f"{MOCK_API_URL}/api/collections/{seqId}",
        json={"id": seqId, "title": "whatever"},
    )
    requests_mock.post(
        f"{MOCK_API_URL}/api/collections",
        json={"id": seqId},
        headers={"Location": f"{MOCK_API_URL}/api/collections/{seqId}"},
    )
    requests_mock.post(
        f"{MOCK_API_URL}/api/collections/{seqId}/items",
        json={"type": "Feature", "id": picId1},
        headers={"Location": f"{MOCK_API_URL}/api/collections/{seqId}/items/{picId1}"},
    )
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles), geovisio=gvsMock, title=None, pictureUploadTimeout=20
    )

    # Check previous pictures are OK
    assert len(uploadReports) == 1
    uploadReport = uploadReports[0]
    assert len(uploadReport.uploaded_pictures) == 1
    assert len(uploadReport.errors) == 0

    # Make other pictures available
    os.rename(datafiles + "/e2.bak", datafiles + "/e2.jpg")
    os.rename(datafiles + "/e3.bak", datafiles + "/e3.jpg")
    with open(datafiles + "/_geovisio.toml", "rb") as f:
        seq = tomli.load(f)
        seq["1"]["pictures"]["e2.jpg"] = {
            "path": str(datafiles) + "/e2.jpg",
            "position": 2,
        }
        seq["1"]["pictures"]["e3.jpg"] = {
            "path": str(datafiles) + "/e3.jpg",
            "position": 3,
        }
        f.close()

    with open(datafiles + "/_geovisio.toml", "wb") as f2:
        tomli_w.dump(seq, f2)
        f2.close()

        # Mock item call to fail
        requests_mock.post(
            f"{MOCK_API_URL}/api/collections/{seqId}/items",
            [
                {
                    "json": {"type": "Feature", "id": picId2},
                    "status_code": 202,
                    "headers": {
                        "Location": f"{MOCK_API_URL}/api/collections/{seqId}/items/{picId2}"
                    },
                },
                {"status_code": 500},
            ],
        )
        uploadReports2 = geovisio_cli.sequences.upload.upload(
            path=Path(datafiles), geovisio=gvsMock, title=None, pictureUploadTimeout=20
        )

        assert len(uploadReports2) == 1
        uploadReport2 = uploadReports2[0]
        assert len(uploadReport2.uploaded_pictures) == 1
        assert len(uploadReport2.errors) == 1


@pytest.mark.parametrize(
    ("picture", "result"),
    (
        (model.Picture(overridden_metadata=None), {}),
        (
            model.Picture(overridden_metadata=PartialGeoPicTags()),
            {},
        ),
        (
            model.Picture(
                overridden_metadata=PartialGeoPicTags(lon=42, model="MAKE")
            ),  # api does not handle overriding model for the moment so it's not in the end result
            {"override_longitude": 42},
        ),
        (
            model.Picture(
                overridden_metadata=PartialGeoPicTags(lat=12, type="flat", ts=424242)
            ),  # api does not handle overriding type for the moment so it's not in the end result
            {
                "override_latitude": 12,
                "override_capture_time": "1970-01-05T21:50:42+00:00",
            },
        ),
        (
            model.Picture(
                overridden_metadata=PartialGeoPicTags(
                    exif={"Exif.Image.Software": "Hugin", "Xmp.xmp.Rating": "5"}
                )
            ),
            {"override_Exif.Image.Software": "Hugin", "override_Xmp.xmp.Rating": "5"},
        ),
    ),
)
def test_get_overriden_metadata(picture, result):
    assert geovisio_cli.sequences.upload._get_overriden_metadata(picture) == result



import pytest
import os
from geovisio_cli import exception, model
import geovisio_cli.sequences.process.standard
import geovisio_cli.sequences.upload
from .conftest import FIXTURE_DIR
from pathlib import Path
from geopic_tag_reader import reader


def test_SplitParams_is_split_needed():
    sp = geovisio_cli.sequences.process.standard.SplitParams()
    assert not sp.is_split_needed()
    sp = geovisio_cli.sequences.process.standard.SplitParams(maxDistance=1)
    assert sp.is_split_needed()
    sp = geovisio_cli.sequences.process.standard.SplitParams(maxTime=1)
    assert sp.is_split_needed()
    sp = geovisio_cli.sequences.process.standard.SplitParams(maxDistance=1, maxTime=1)
    assert sp.is_split_needed()


@pytest.mark.parametrize(
    ("data", "method", "expected"),
    (
        (["1.jpg", "2.jpg", "3.jpg"], "filename-asc", ["1.jpg", "2.jpg", "3.jpg"]),
        (["3.jpg", "1.jpg", "2.jpg"], "filename-asc", ["1.jpg", "2.jpg", "3.jpg"]),
        (["3.jpg", "1.jpg", "2.jpeg"], "filename-asc", ["1.jpg", "2.jpeg", "3.jpg"]),
        (["10.jpg", "5.jpg", "1.jpg"], "filename-asc", ["1.jpg", "5.jpg", "10.jpg"]),
        (["C.jpg", "A.jpg", "B.jpg"], "filename-asc", ["A.jpg", "B.jpg", "C.jpg"]),
        (
            ["CAM1_001.jpg", "CAM2_002.jpg", "CAM1_002.jpg"],
            "filename-asc",
            ["CAM1_001.jpg", "CAM1_002.jpg", "CAM2_002.jpg"],
        ),
        (["1.jpg", "2.jpg", "3.jpg"], "filename-desc", ["3.jpg", "2.jpg", "1.jpg"]),
        (["3.jpg", "1.jpg", "2.jpg"], "filename-desc", ["3.jpg", "2.jpg", "1.jpg"]),
        (["3.jpg", "1.jpg", "2.jpeg"], "filename-desc", ["3.jpg", "2.jpeg", "1.jpg"]),
        (["10.jpg", "5.jpg", "1.jpg"], "filename-desc", ["10.jpg", "5.jpg", "1.jpg"]),
        (["C.jpg", "A.jpg", "B.jpg"], "filename-desc", ["C.jpg", "B.jpg", "A.jpg"]),
        (
            ["CAM1_001.jpg", "CAM2_002.jpg", "CAM1_002.jpg"],
            "filename-desc",
            ["CAM2_002.jpg", "CAM1_002.jpg", "CAM1_001.jpg"],
        ),
    ),
)
def test_sort_files_names(data, method, expected):
    dataPictures = [model.Picture(path=p) for p in data]
    resPictures = geovisio_cli.sequences.process.standard._sort_files(
        dataPictures, method
    )
    assert expected == [pic.path for pic in resPictures]


@pytest.mark.parametrize(
    ("data", "method", "expected"),
    (
        (
            [["1.jpg", 1], ["2.jpg", 2], ["3.jpg", 3]],
            "time-asc",
            ["1.jpg", "2.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 2], ["2.jpg", 3], ["3.jpg", 1]],
            "time-asc",
            ["3.jpg", "1.jpg", "2.jpg"],
        ),
        (
            [["1.jpg", 1.01], ["2.jpg", 1.02], ["3.jpg", 1.03]],
            "time-asc",
            ["1.jpg", "2.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 1], ["2.jpg", 2], ["3.jpg", 3]],
            "time-desc",
            ["3.jpg", "2.jpg", "1.jpg"],
        ),
        (
            [["1.jpg", 2], ["2.jpg", 3], ["3.jpg", 1]],
            "time-desc",
            ["2.jpg", "1.jpg", "3.jpg"],
        ),
        (
            [["1.jpg", 1.01], ["2.jpg", 1.02], ["3.jpg", 1.03]],
            "time-desc",
            ["3.jpg", "2.jpg", "1.jpg"],
        ),
    ),
)
def test_sort_files_time(data, method, expected):
    dataPictures = []
    for p in data:
        name, ts = p
        m = reader.GeoPicTags(
            lon=47.7,
            lat=-1.78,
            ts=ts,
            heading=0,
            type="flat",
            make="Panoramax",
            model="180++",
            focal_length=4,
            crop=None,
        )
        dataPictures.append(model.Picture(path=name, metadata=m))

    resPictures = geovisio_cli.sequences.process.standard._sort_files(
        dataPictures, method
    )
    assert expected == [pic.path for pic in resPictures]


@pytest.mark.parametrize(
    ("pics", "maxTime", "maxDist"),
    (
        [  # Single sequence
            [
                {
                    "lat": 48.0000001,
                    "lon": -1.7800001,
                    "ts": 1,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.0000002,
                    "lon": -1.7800002,
                    "ts": 2,
                    "heading": 100,
                    "seq": 0,
                },
            ],
            None,
            None,
        ],
        [  # Time excedeed
            [
                {
                    "lat": 48.0000001,
                    "lon": -1.7800001,
                    "ts": 1,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.0000002,
                    "lon": -1.7800002,
                    "ts": 500,
                    "heading": 100,
                    "seq": 1,
                },
            ],
            10,
            None,
        ],
        [  # Time excedeed, reverse
            [
                {
                    "lat": 48.0000001,
                    "lon": -1.7800001,
                    "ts": 500,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.0000002,
                    "lon": -1.7800002,
                    "ts": 1,
                    "heading": 100,
                    "seq": 1,
                },
            ],
            10,
            None,
        ],
        [  # Distance excedeed
            [
                {
                    "lat": 48.0000001,
                    "lon": -1.7800001,
                    "ts": 1,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.1000000,
                    "lon": -1.7800002,
                    "ts": 2,
                    "heading": 100,
                    "seq": 1,
                },
            ],
            None,
            1,
        ],
        [  # Many sequences
            [
                {
                    "lat": 48.0000001,
                    "lon": -1.7800001,
                    "ts": 1,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.0000002,
                    "lon": -1.7800002,
                    "ts": 2,
                    "heading": 100,
                    "seq": 0,
                },
                {
                    "lat": 48.0000003,
                    "lon": -1.7800003,
                    "ts": 3,
                    "heading": 100,
                    "seq": 0,
                },
                # Distance excedeed
                {
                    "lat": 48.1000000,
                    "lon": -1.7800001,
                    "ts": 4,
                    "heading": 100,
                    "seq": 1,
                },
                {
                    "lat": 48.1000001,
                    "lon": -1.7800001,
                    "ts": 5,
                    "heading": 100,
                    "seq": 1,
                },
                {
                    "lat": 48.1000002,
                    "lon": -1.7800001,
                    "ts": 6,
                    "heading": 100,
                    "seq": 1,
                },
                # Time excedeed
                {
                    "lat": 48.1000003,
                    "lon": -1.7800001,
                    "ts": 100,
                    "heading": 100,
                    "seq": 2,
                },
                {
                    "lat": 48.1000004,
                    "lon": -1.7800001,
                    "ts": 101,
                    "heading": 100,
                    "seq": 2,
                },
                {
                    "lat": 48.1000005,
                    "lon": -1.7800001,
                    "ts": 102,
                    "heading": 100,
                    "seq": 2,
                },
            ],
            30,
            100,
        ],
    ),
)
def test_split_pictures_into_sequences(pics, maxTime, maxDist):
    sp = geovisio_cli.sequences.process.standard.SplitParams(
        maxDistance=maxDist, maxTime=maxTime
    )
    inputPics = []
    expectedPics = [[]]

    for id, pic in enumerate(pics):
        inputPics.append(
            model.Picture(
                id=f"{id}",
                metadata=reader.GeoPicTags(
                    lat=pic["lat"],
                    lon=pic["lon"],
                    ts=pic["ts"],
                    heading=pic["heading"],
                    type="equirectangular",
                    make=None,
                    model=None,
                    focal_length=None,
                    crop=None,
                ),
            )
        )

        if len(expectedPics) - 1 < pic["seq"]:
            expectedPics.append([])
        expectedPics[pic["seq"]].append(f"{id}")

    res = geovisio_cli.sequences.process.standard._split_pictures_into_sequences(
        inputPics, "some_file.toml", sp
    )
    print("Got     ", [[p.id for p in r.pictures] for r in res.sequences])
    print("Expected", expectedPics)
    assert len(res.sequences) == len(expectedPics)

    for i, resSubSeq in enumerate(res.sequences):
        print("Checking sequence", i)
        assert len(resSubSeq.pictures) == len(expectedPics[i])
        for j, resSubSeqPic in enumerate(resSubSeq.pictures):
            print(" -> Checking pic", j)
            assert resSubSeqPic.id == expectedPics[i][j]


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequences(datafiles):
    # First read : position is based on picture names
    seqs = geovisio_cli.sequences.process.standard._read_sequences(Path(datafiles))
    seqsTomlPath = os.path.join(datafiles, model.SEQUENCE_TOML_FILE)
    seqs.persist()

    assert os.path.isfile(seqsTomlPath)

    # Edit TOML file : position is inverted
    with open(seqsTomlPath, "r+") as seqsToml:
        editedSeqsToml = seqsToml.read()
        editedSeqsToml = (
            editedSeqsToml.replace("position = 1", "position = A")
            .replace("position = 2", "position = 1")
            .replace("position = A", "position = 2")
        )
        seqsToml.seek(0)
        seqsToml.write(editedSeqsToml)
        seqsToml.close()

        # Read sequence 2 : position should match edited TOML
        seqs = geovisio_cli.sequences.process.standard._read_sequences(Path(datafiles))
        seq = seqs.sequences[0]
        assert seq.pictures[0].path.endswith("e2.jpg")
        assert seq.pictures[1].path.endswith("e1.jpg")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
    os.path.join(FIXTURE_DIR, "invalid_pic.jpg"),
)
def test_read_sequences_invalid_file(datafiles):
    # Read sequence from files
    seqs = geovisio_cli.sequences.process.standard._read_sequences(Path(datafiles))
    seqs.persist()

    # Check if invalid_pic is marked as broken
    seq2 = model.ManySequences.read_from_file(seqs.toml_file).sequences[0]
    assert len(seq2.pictures) == 4
    assert seq2.pictures[0].status == "broken-metadata"
    assert seq2.pictures[1].status is None
    assert seq2.pictures[2].status is None
    assert seq2.pictures[3].status is None


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequences_sort_method_changed_set2unset(datafiles):
    # Write toml with sort method defined
    seq = geovisio_cli.sequences.process.standard._read_sequences(
        Path(datafiles), sortMethod=model.SortMethod.time_desc
    )
    seqTomlPath = os.path.join(datafiles, model.SEQUENCE_TOML_FILE)
    seq.persist()

    # Read sequence from toml without sort method = should reuse read one
    seq = geovisio_cli.sequences.process.standard._read_sequences(Path(datafiles))
    assert seq.sequences[0].pictures[0].path.endswith("e2.jpg")
    assert seq.sequences[0].pictures[1].path.endswith("e1.jpg")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
)
def test_read_sequences_sort_method_changed_different(datafiles):
    # Write toml with sort method defined
    seq = geovisio_cli.sequences.process.standard._read_sequences(
        Path(datafiles), sortMethod=model.SortMethod.time_desc
    )
    seqTomlPath = os.path.join(datafiles, model.SEQUENCE_TOML_FILE)
    seq.persist()

    # Read sequence from toml without sort method = should reuse read one
    with pytest.raises(exception.CliException) as e:
        seq = geovisio_cli.sequences.process.standard._read_sequences(
            Path(datafiles), sortMethod=model.SortMethod.filename_asc
        )

    assert e.match("Sort method passed as argument")



from dataclasses import dataclass, field, fields
from enum import Enum
import os
from typing import List, Optional, Dict, Any, Tuple
from geopic_tag_reader import reader
from . import utils
import tomli  # type: ignore
import tomli_w  # type: ignore
from pathlib import Path
from haversine import Unit, haversine  # type: ignore


@dataclass
class Geovisio:
    url: str
    token: Optional[str] = None


BROKEN_METADATA_STATUS = "broken-metadata"
BROKEN_UPLOAD_STATUS = "broken"
DUPLICATE_STATUS = "duplicate"


@dataclass
class Picture:
    path: Optional[str] = None
    id: Optional[str] = None
    location: Optional[str] = None
    status: Optional[str] = None
    metadata: Optional[reader.GeoPicTags] = None
    overridden_metadata: Optional[reader.PartialGeoPicTags] = None

    def toml(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {
            "path": self.path,
            "id": self.id,
            "location": self.location,
            "status": self.status,
        }
        if self.overridden_metadata:
            d["overriden_metadata"] = {}
            for field in fields(self.overridden_metadata):
                v = getattr(self.overridden_metadata, field.name)
                if v:
                    d["overriden_metadata"][field.name] = v
        return utils.removeNoneInDict(d)

    @staticmethod
    def from_toml(data):
        overriden_metadata = None
        if data.get("overriden_metadata"):
            overriden_metadata = reader.PartialGeoPicTags(
                **data.get("overriden_metadata")
            )
        return Picture(
            path=data.get("path"),
            id=data.get("id"),
            location=data.get("location"),
            status=data.get("status"),
            overridden_metadata=overriden_metadata,
        )

    def has_mandatory_metadata(self):
        """To be valid a picture should have a coordinate and a timestamp"""
        if self.metadata is not None:
            return True
        mandatory_fields = ["lon", "lat", "ts"]
        for m in mandatory_fields:
            if getattr(self.overridden_metadata, m) is None:
                return False
        return True

    def update_overriden_metadata(self, new_metadata: reader.PartialGeoPicTags):
        """Update overriden metadata with new value only if there is no existing value"""
        if self.overridden_metadata is None:
            self.overridden_metadata = new_metadata
            return
        for field in fields(self.overridden_metadata):
            old_value = getattr(self.overridden_metadata, field.name)
            if old_value is None:
                setattr(
                    self.overridden_metadata,
                    field.name,
                    getattr(new_metadata, field.name),
                )


@dataclass
class Split:
    prevPic: Picture
    nextPic: Picture
    reason: str


@dataclass
class InteriorOrientation:
    make: str
    model: str
    field_of_view: Optional[int]


class SortMethod(str, Enum):
    filename_asc = "filename-asc"
    filename_desc = "filename-desc"
    time_asc = "time-asc"
    time_desc = "time-desc"


@dataclass
class Sequence:
    title: str = ""
    path: str = ""
    id: Optional[str] = None
    location: Optional[str] = None
    producer: Optional[str] = None
    interior_orientation: List[InteriorOrientation] = field(default_factory=lambda: [])
    pictures: List[Picture] = field(default_factory=lambda: [])
    sort_method: Optional[SortMethod] = None

    def toml(self):
        res = {
            "sequence": utils.removeNoneInDict(
                {
                    "title": self.title,
                    "path": self.path,
                    "id": self.id,
                    "location": self.location,
                    "producer": self.producer,
                    "sort_method": self.sort_method.value
                    if self.sort_method is not None
                    else None,
                }
            ),
            "pictures": {},
        }

        for pos, pic in enumerate(self.pictures, start=1):
            pict = pic.toml()
            pict["position"] = pos
            if pic.path:
                res["pictures"][os.path.basename(pic.path)] = pict

        return res

    @staticmethod
    def from_toml(data):
        s = Sequence()
        s.update_from_toml(data)
        return s

    def update_from_toml(self, data):
        if data.get("sequence"):
            self.title = data["sequence"].get("title", "")
            self.path = data["sequence"].get("path", "")
            self.id = data["sequence"].get("id")
            self.location = data["sequence"].get("location")
            self.producer = data["sequence"].get("producer")
            self.sort_method = (
                SortMethod(data["sequence"]["sort_method"])
                if "sort_method" in data["sequence"]
                else None
            )

        if data.get("pictures"):
            self.pictures = [
                Picture.from_toml(picData)
                for picId, picData in sorted(
                    data["pictures"].items(),
                    key=lambda item: int(item[1].get("position") or -1),
                )
            ]

    def find_duplicates(
        self,
        maxDistance: Optional[float] = None,
        maxRotationAngle: Optional[int] = None,
    ) -> None:
        """Mark pictures too close to each other as duplicates.

        This avoids to upload similar pictures.

        Parameters
        ----------
        maxDistance : float
            The maximum distance to consider two pictures as duplicates (in meters)
        maxRotationAngle : int
            The maximum rotation angle to consider two pictures as duplicated (in degrees)
        """

        if maxDistance is None:
            return None

        lastNonDuplicatedPicId = 0

        for i, currentPic in enumerate(self.pictures):
            if i == 0:
                continue

            prevPic = self.pictures[lastNonDuplicatedPicId]

            if prevPic.metadata is None or currentPic.metadata is None:
                continue

            # Compare distance
            dist = haversine(
                (prevPic.metadata.lat, prevPic.metadata.lon),
                (currentPic.metadata.lat, currentPic.metadata.lon),
                unit=Unit.METERS,
            )

            if maxDistance is not None and dist <= maxDistance:
                # Compare angle (if available on both images)
                if (
                    maxRotationAngle is not None
                    and prevPic.metadata.heading is not None
                    and currentPic.metadata.heading is not None
                ):
                    deltaAngle = abs(
                        currentPic.metadata.heading - prevPic.metadata.heading
                    )

                    if deltaAngle <= maxRotationAngle:
                        currentPic.status = DUPLICATE_STATUS
                    else:
                        lastNonDuplicatedPicId = i
                else:
                    currentPic.status = DUPLICATE_STATUS
            else:
                lastNonDuplicatedPicId = i

    def all_done(self):
        return self.nb_waiting() + self.nb_preparing() == 0

    def nb_ready(self):
        return sum((1 for p in self.pictures if p.status == "ready"))

    def nb_waiting(self):
        return sum((1 for p in self.pictures if p.status == "waiting-for-process"))

    def nb_duplicates(self):
        return sum((1 for p in self.pictures if p.status == DUPLICATE_STATUS))

    def nb_preparing(self):
        return sum(
            (
                1
                for p in self.pictures
                if (p.status and p.status.startswith("preparing"))
            )
        )

    def nb_broken(self):
        return sum((1 for p in self.pictures if p.status == BROKEN_UPLOAD_STATUS))


@dataclass
class ManySequences:
    """ManySequences is a single handler for multiple sequences found in a single folder."""

    toml_file: Path
    sequences: List[Sequence] = field(default_factory=lambda: [])
    splits: List[Split] = field(default_factory=lambda: [])

    def toml(self):
        res = {}
        for i, s in enumerate(self.sequences, start=1):
            res[str(i)] = s.toml()
        return res

    def has_same_sort_method(self, sortMethod):
        if sortMethod is None:
            return True
        else:
            for s in self.sequences:
                if s.sort_method != sortMethod:
                    return False
            return True

    def is_empty(self):
        if len(self.sequences) == 0:
            return True
        return not any((len(s.pictures) != 0 for s in self.sequences))

    def has_valid_pictures(self):
        for s in self.sequences:
            for p in s.pictures:
                if p.status != BROKEN_METADATA_STATUS:
                    return True
        return False

    @classmethod
    def read_from_file(cls, file: Path):
        """Read multiple sequences from a single TOML config file"""

        with open(file, "rb") as f:
            tomlContent = tomli.load(f)
            f.close()

            # Retro-compatibility with old file formats (single-sequence in file)
            if tomlContent.get("sequence") is not None:
                ms = ManySequences(
                    toml_file=file, sequences=[Sequence.from_toml(tomlContent)]
                )
            else:
                ms = ManySequences(
                    toml_file=file,
                    sequences=[
                        Sequence.from_toml(seqData) for seqData in tomlContent.values()
                    ],
                )

            return ms

    def persist(self) -> Path:
        """Writes TOML sequence metadata file"""

        with open(self.toml_file, "wb") as f:
            tomli_w.dump(self.toml(), f)
            f.close()

        return self.toml_file


SEQUENCE_TOML_FILE = "_geovisio.toml"



from pathlib import Path
from geovisio_cli import model
import pytest
from geopic_tag_reader.reader import PartialGeoPicTags, GeoPicTags


def test_Picture_toml():
    p = model.Picture(
        path="/tmp/1.jpg",
        id="blerg",
        location="http://api.geovisio.fr/blerg",
        status="ready",
    )
    assert p == model.Picture.from_toml(p.toml())


def test_Picture_with_overriden_metadata_toml():
    p = model.Picture(
        path="/tmp/1.jpg",
        id="blerg",
        location="http://api.geovisio.fr/blerg",
        status="ready",
        overridden_metadata=PartialGeoPicTags(lat=42, lon=12),
    )
    assert p == model.Picture.from_toml(p.toml())


def test_Sequence_toml():
    pics = [
        model.Picture(
            path="/tmp/bla/1.jpg",
            id="bla1",
            location="http://api.geovisio.fr/bla/1",
            status="ready",
        ),
        model.Picture(
            path="/tmp/bla/2.jpg",
            id="bla2",
            location="http://api.geovisio.fr/bla/2",
            status="ready",
        ),
    ]
    s = model.Sequence(
        title="Bla",
        path="/tmp/bla",
        id="blabla",
        location="http://api.geovisio.fr/bla",
        producer="BBC (Bla Bla Corporation)",
        pictures=pics,
        sort_method=model.SortMethod.time_desc,
    )
    assert s == model.Sequence.from_toml(s.toml())


def test_ManySequences_toml(tmpdir):
    pics1 = [
        model.Picture(
            path="/tmp/bla/1.jpg",
            id="bla1",
            location="http://api.geovisio.fr/bla/1",
            status="ready",
        ),
        model.Picture(
            path="/tmp/bla/2.jpg",
            id="bla2",
            location="http://api.geovisio.fr/bla/2",
            status="ready",
        ),
    ]
    seq1 = model.Sequence(
        title="Bla",
        path="/tmp/bla",
        id="blabla",
        location="http://api.geovisio.fr/bla",
        producer="BBC (Bla Bla Corporation)",
        pictures=pics1,
        sort_method=model.SortMethod.time_desc,
    )
    pics2 = [
        model.Picture(
            path="/tmp/bla2/1.jpg",
            id="bla2-1",
            location="http://api.geovisio.fr/bla2/1",
            status="ready",
        ),
        model.Picture(
            path="/tmp/bla2/2.jpg",
            id="bla2-2",
            location="http://api.geovisio.fr/bla2/2",
            status="ready",
        ),
    ]
    seq2 = model.Sequence(
        title="Bla2",
        path="/tmp/bla2",
        id="blabla2",
        location="http://api.geovisio.fr/bla2",
        producer="BBC (Bla Bla Corporation)",
        pictures=pics2,
        sort_method=model.SortMethod.time_asc,
    )
    toml_file = tmpdir / model.SEQUENCE_TOML_FILE
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1, seq2])
    ms.persist()
    assert ms == model.ManySequences.read_from_file(toml_file)


def test_ManySequences_has_same_sort_method():
    toml_file = Path("some_file.toml")
    seq1 = model.Sequence(sort_method=model.SortMethod.time_asc)
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1])

    assert ms.has_same_sort_method(None)
    assert ms.has_same_sort_method(model.SortMethod.time_asc)
    assert not ms.has_same_sort_method(model.SortMethod.time_desc)

    ms = model.ManySequences(toml_file=toml_file, sequences=[])
    assert ms.has_same_sort_method(None)
    assert ms.has_same_sort_method(model.SortMethod.time_asc)


def test_ManySequences_is_empty():
    toml_file = Path("some_file.toml")
    ms = model.ManySequences(toml_file=toml_file, sequences=[])
    assert ms.is_empty()

    seq1 = model.Sequence()
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1])
    assert ms.is_empty()

    pic1 = model.Picture(path="/tmp/bla/1.jpg")
    seq1 = model.Sequence(pictures=[pic1])
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1])
    assert not ms.is_empty()


def test_ManySequences_has_valid_pictures():
    toml_file = Path("some_file.toml")
    ms = model.ManySequences(toml_file=toml_file, sequences=[])
    assert not ms.has_valid_pictures()

    pic1 = model.Picture(path="/tmp/bla/1.jpg", status="broken-metadata")
    seq1 = model.Sequence(pictures=[pic1])
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1])
    assert not ms.has_valid_pictures()

    pic1 = model.Picture(path="/tmp/bla/1.jpg")
    seq1 = model.Sequence(pictures=[pic1])
    ms = model.ManySequences(toml_file=toml_file, sequences=[seq1])
    assert ms.has_valid_pictures()


def test_rw_sequences_toml(tmp_path):
    s = model.Sequence(
        title="SEQUENCE",
        id="blab-blabla-blablabla",
        path=str(tmp_path),
        pictures=[
            model.Picture(
                id="blou-bloublou-bloubloublou-1", path=str(tmp_path / "1.jpg")
            ),
            model.Picture(
                id="blou-bloublou-bloubloublou-2", path=str(tmp_path / "2.jpg")
            ),
            model.Picture(
                id="blou-bloublou-bloubloublou-3", path=str(tmp_path / "3.jpg")
            ),
        ],
        sort_method=model.SortMethod.time_desc,
    )
    tomlFile = tmp_path / "_geovisio.toml"
    ms = model.ManySequences(tomlFile, sequences=[s])
    res = ms.persist()
    assert res == tomlFile

    res2 = model.ManySequences.read_from_file(tomlFile)
    assert ms == res2


@pytest.mark.parametrize(
    ("picture", "is_ok"),
    (
        (
            model.Picture(
                overridden_metadata=PartialGeoPicTags(
                    lon=12.2, lat=43.4, ts=1516589529.9
                )
            ),
            True,
        ),
        (
            model.Picture(overridden_metadata=PartialGeoPicTags(lon=12.2, lat=43.4)),
            False,
        ),
        (
            model.Picture(
                overridden_metadata=PartialGeoPicTags(lon=12.2, ts=1516589529.9)
            ),
            False,
        ),
        (
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.2,
                    lat=43.4,
                    ts=1516589529.9,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=None,
                    make=None,
                ),
                overridden_metadata=PartialGeoPicTags(),
            ),
            True,
        ),
    ),
)
def test_has_mandatory_metadata(picture, is_ok):
    assert picture.has_mandatory_metadata() == is_ok


def test_update_overriden_metadata():
    pic = model.Picture(
        overridden_metadata=PartialGeoPicTags(
            lon=12.2, ts=12, make="CANON", type="flat"
        )
    )

    pic.update_overriden_metadata(
        PartialGeoPicTags(lat=43.4, ts=4242, model="Some model", type="flat")
    )

    # after override, fields should not have been changed if set initially
    assert pic.overridden_metadata == PartialGeoPicTags(
        lon=12.2, lat=43.4, ts=12, make="CANON", model="Some model", type="flat"
    )


def test_Sequence_find_duplicates():
    def getData():
        pictures = [
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.2,
                    lat=43.4,
                    ts=1516589529,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=12,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.2000001,
                    lat=43.4000001,
                    ts=1516589530,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=12,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.2000002,
                    lat=43.4000002,
                    ts=1516589531,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=102,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.3,
                    lat=43.5,
                    ts=1516589532,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=102,
                    make=None,
                )
            ),
        ]
        s = model.Sequence(pictures=pictures)
        return s

    # Using both dist and angle
    s = getData()
    s.find_duplicates(maxDistance=5, maxRotationAngle=10)
    assert s.pictures[0].status == None
    assert s.pictures[1].status == model.DUPLICATE_STATUS
    assert s.pictures[2].status == None
    assert s.pictures[3].status == None

    # Using only distance
    s = getData()
    s.find_duplicates(maxDistance=5)
    assert s.pictures[0].status == None
    assert s.pictures[1].status == model.DUPLICATE_STATUS
    assert s.pictures[2].status == model.DUPLICATE_STATUS
    assert s.pictures[3].status == None

    # Using no param
    s = getData()
    s.find_duplicates()
    assert s.pictures[0].status == None
    assert s.pictures[1].status == None
    assert s.pictures[2].status == None
    assert s.pictures[3].status == None

    # Using only rotation
    s = getData()
    s.find_duplicates(maxRotationAngle=5)
    assert s.pictures[0].status == None
    assert s.pictures[1].status == None
    assert s.pictures[2].status == None
    assert s.pictures[3].status == None


def test_Sequence_find_duplicates_many_near():
    def getData():
        pictures = [
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.2,
                    lat=43.4,
                    ts=1516589529,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=12,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.20001,
                    lat=43.40001,
                    ts=1516589530,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=12,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.20002,
                    lat=43.40002,
                    ts=1516589531,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=102,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.20003,
                    lat=43.40003,
                    ts=1516589532,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=102,
                    make=None,
                )
            ),
            model.Picture(
                metadata=GeoPicTags(
                    lon=12.20004,
                    lat=43.40004,
                    ts=1516589532,
                    type="flat",
                    model=None,
                    crop=None,
                    focal_length=None,
                    heading=102,
                    make=None,
                )
            ),
        ]
        s = model.Sequence(pictures=pictures)
        return s

    # 1m
    s = getData()
    s.find_duplicates(maxDistance=1)
    assert s.pictures[0].status == None
    assert s.pictures[1].status == None
    assert s.pictures[2].status == None
    assert s.pictures[3].status == None
    assert s.pictures[4].status == None

    # 2m
    s = getData()
    s.find_duplicates(maxDistance=2)
    assert s.pictures[0].status == None
    assert s.pictures[1].status == model.DUPLICATE_STATUS
    assert s.pictures[2].status == None
    assert s.pictures[3].status == model.DUPLICATE_STATUS
    assert s.pictures[4].status == None



from dataclasses import dataclass
from haversine import Unit, haversine  # type: ignore
from geovisio_cli import exception
from geovisio_cli.exception import CliException
from geovisio_cli.model import (
    ManySequences,
    Picture,
    Sequence,
    SortMethod,
    Split,
    SEQUENCE_TOML_FILE,
    BROKEN_METADATA_STATUS,
    BROKEN_UPLOAD_STATUS,
)
from geovisio_cli.sequences import external_metadata
from geopic_tag_reader import reader
from rich import print
from rich.progress import track
import logging
import os
from pathlib import Path, PurePath
from typing import List, Optional


@dataclass
class MergeParams:
    maxDistance: Optional[float] = None
    maxRotationAngle: Optional[int] = None

    def is_merge_needed(self):
        # Only check max distance, as max rotation angle is only useful when dist is defined
        return self.maxDistance is not None


@dataclass
class SplitParams:
    maxDistance: Optional[int] = None
    maxTime: Optional[int] = None

    def is_split_needed(self):
        return self.maxDistance is not None or self.maxTime is not None


def _split_pictures_into_sequences(
    pictures: List[Picture],
    path: str,
    splitParams: Optional[SplitParams] = SplitParams(),
    title: Optional[str] = None,
    sort_method: Optional[SortMethod] = None,
) -> ManySequences:
    """Split a list of pictures into multiple lists of pictures (to create sequences)
    based on maximum distance, time and rotation between two pictures.
    Note that this function expect pictures to be sorted and have their metadata set.
    Parameters
    ----------
    pictures : Picture[]
        List of pictures to check, sorted and with metadata defined
    splitParams : SplitParams
        The parameters to define deltas between two pictures
    title : str
        Title to set on resulting sequences
    path : str
        Folder to set on resulting sequences
    sort_method : SortMethod
        Sort method to set on resulting sequences
    Returns
    -------
    ManySequences
        List of pictures splitted into smaller sequences
    """

    stoml = os.path.join(path, SEQUENCE_TOML_FILE)
    ms = ManySequences(toml_file=Path(stoml))

    # No split parameters given -> just return given pictures
    if splitParams is None or not splitParams.is_split_needed():
        ms.sequences = [
            Sequence(
                title=title or "",
                path=path,
                pictures=pictures,
                sort_method=sort_method,
            )
        ]
        return ms

    currentPicList: List[Picture] = []

    for pic in pictures:
        if len(currentPicList) == 0:  # No checks for 1st pic
            currentPicList.append(pic)
        else:
            lastPic = currentPicList[-1]

            # Missing metadata -> skip
            if lastPic.metadata is None or pic.metadata is None:
                currentPicList.append(pic)
                continue

            # Time delta
            timeOutOfDelta = (
                False
                if splitParams.maxTime is None
                else abs(lastPic.metadata.ts - pic.metadata.ts) > splitParams.maxTime
            )

            # Distance delta
            distance = haversine(
                (lastPic.metadata.lat, lastPic.metadata.lon),
                (pic.metadata.lat, pic.metadata.lon),
                unit=Unit.METERS,
            )
            distanceOutOfDelta = (
                False
                if splitParams.maxDistance is None
                else distance > splitParams.maxDistance
            )

            # One of deltas maxed -> create new sequence
            if timeOutOfDelta or distanceOutOfDelta:
                # Mark the reason for split
                picRangeTxt = f"between {PurePath(lastPic.path or '').stem} and {PurePath(pic.path or '').stem}"
                reason = ""
                if timeOutOfDelta:
                    reason = f"Too much time {picRangeTxt} ({round(abs(lastPic.metadata.ts - pic.metadata.ts))} seconds)"
                elif distanceOutOfDelta:
                    reason = (
                        f"Too much distance {picRangeTxt} ({round(distance)} meters)"
                    )

                ms.splits.append(Split(lastPic, pic, reason))
                ms.sequences.append(
                    Sequence(
                        title=title or "",
                        path=path,
                        pictures=currentPicList,
                        sort_method=sort_method,
                    )
                )
                currentPicList = [pic]

            # Otherwise, still in same sequence
            else:
                currentPicList.append(pic)

    ms.sequences.append(
        Sequence(
            title=title or "",
            path=path,
            pictures=currentPicList,
            sort_method=sort_method,
        )
    )

    return ms


def _sort_files(
    pictures: List[Picture], method: Optional[SortMethod] = SortMethod.time_asc
) -> List[Picture]:
    """Sorts pictures according to their file name
    Parameters
    ----------
    pictures : Picture[]
        List of pictures to sort
    method : SortMethod
        Sort logic to adopt (time-asc, time-desc, filename-asc, filename-desc)
    Returns
    -------
    Picture[]
        List of pictures, sorted
    """

    if method is None:
        method = SortMethod.time_asc

    if method not in [item.value for item in SortMethod]:
        raise exception.CliException("Invalid sort strategy: " + str(method))

    # Get the sort logic
    strat, order = method.split("-")

    # Sort based on filename
    if strat == "filename":
        # Check if pictures can be sorted by numeric notation
        hasNonNumber = False
        for p in pictures:
            try:
                int(PurePath(p.path or "").stem)
            except:
                hasNonNumber = True
                break

        def sort_fct(p):
            return (
                PurePath(p.path or "").stem
                if hasNonNumber
                else int(PurePath(p.path or "").stem)
            )

        pictures.sort(key=sort_fct)

    # Sort based on picture ts
    elif strat == "time":
        pictures.sort(key=lambda p: p.metadata.ts if p.metadata is not None else 0)

    if order == "desc":
        pictures.reverse()

    return pictures


def _search_for_pictures(path: Path):
    """
    Collect all the pictures in a directory
    """
    files = []
    for f in path.iterdir():
        if not f.is_file():
            continue
        if f.suffix.lower() not in [".jpg", ".jpeg"]:
            continue
        files.append(f)
    return files


def _load_pictures(
    picture_files: List[Path],
    metadata_handler: Optional[external_metadata.MetadataHandler],
) -> List[Picture]:
    """Read a list of pictures and their associated metadata from a list of files."""
    pictures = []
    for f in track(picture_files, description="🔍 Listing pictures..."):
        picture = Picture(path=str(f))

        if metadata_handler:
            picture.overridden_metadata = metadata_handler.get(f)
        try:
            with open(str(f), "rb") as img:
                meta = reader.readPictureMetadata(img.read())
                picture.metadata = meta

        except reader.PartialExifException as e:
            # override picture metadata with the one found in the exif tags
            # if a tag is found in both exif and external metadata, the external ones are used
            picture.update_overriden_metadata(e.tags)
            if not picture.has_mandatory_metadata():
                picture.status = BROKEN_METADATA_STATUS

        except Exception as e:
            logging.warning(f"Picture {str(f)} has invalid metadata: {str(e)}")
            picture.status = BROKEN_METADATA_STATUS

        pictures.append(picture)

    return pictures


def _read_sequences(
    path: Path,
    title: Optional[str] = None,
    sortMethod: Optional[SortMethod] = None,
    splitParams: Optional[SplitParams] = None,
    mergeParams: Optional[MergeParams] = None,
) -> ManySequences:
    if not path.is_dir():
        raise CliException(f"{path} is not a directory, cannot read pictures")

    if title is None:
        title = path.resolve().name

    sequences = None
    stoml = os.path.join(path, SEQUENCE_TOML_FILE)

    # Check if a TOML file exists, then use it instead of generating one
    if os.path.isfile(stoml):
        try:
            sequences = ManySequences.read_from_file(Path(stoml))
            print(f"📄 Using metadata from existing config file: {stoml}")

            # Check sort method
            if sortMethod is not None and not sequences.has_same_sort_method(
                sortMethod
            ):
                raise CliException(
                    f'Sort method passed as argument ({sortMethod.value}) is different from the one defined in your metadata file.\nYou may either change --sort-method argument from command-line, or change "sort_method" in your metadata file.'
                )
        except EOFError:
            pass

    # Create sequence from pictures files
    if sequences is None:
        # Identify pictures files
        picture_files = _search_for_pictures(path)
        metadata_handler = external_metadata.find_handler(path)

        # Read pictures metadata (with loader)
        pictures = _load_pictures(picture_files, metadata_handler)

        # Sort, split and create sequences
        pictures = _sort_files(pictures, sortMethod)
        sequences = _split_pictures_into_sequences(
            pictures, str(path), splitParams, title=title, sort_method=sortMethod
        )

        # Look for duplicates
        if mergeParams is not None and mergeParams.is_merge_needed():
            for s in sequences.sequences:
                s.find_duplicates(mergeParams.maxDistance, mergeParams.maxRotationAngle)

    if sequences.is_empty():
        raise CliException(
            "❌ No picture was found in given folder.\nMake sure your folder contains at least one valid JPEG file."
        )

    # Check if at least one picture is valid
    if not sequences.has_valid_pictures():
        raise CliException(
            "❌ All read pictures have invalid metadata.\nPlease check if your pictures are geolocated and have a date defined.\nFor more information: https://gitlab.com/geovisio/api/-/blob/develop/docs/15_Pictures_requirements.md"
        )

    return sequences


def process(
    path: Path,
    title: Optional[str],
    sortMethod: Optional[SortMethod] = None,
    splitParams: Optional[SplitParams] = None,
    mergeParams: Optional[MergeParams] = None,
) -> ManySequences:
    ms = _read_sequences(path, title, sortMethod, splitParams, mergeParams)
    ms.persist()

    if (
        splitParams is not None
        and splitParams.is_split_needed()
        and len(ms.sequences) > 1
    ):
        print(f"\n🗂️  Pictures are split into {len(ms.sequences)} sequences")
        for i in range(len(ms.splits)):
            if i == 0:
                d = (
                    ""
                    if ms.sequences[0].nb_duplicates() == 0
                    else f" including {ms.sequences[0].nb_duplicates()} duplicates"
                )
                print(f"  Sequence 1 ({len(ms.sequences[0].pictures)} pictures{d})")
            print(f"    ✂️ {ms.splits[i].reason}")
            d = (
                ""
                if ms.sequences[i + 1].nb_duplicates() == 0
                else f" including {ms.sequences[i+1].nb_duplicates()} duplicates"
            )
            print(f"  Sequence {i+2} ({len(ms.sequences[i+1].pictures)} pictures{d})")
    else:
        d = (
            ""
            if ms.sequences[0].nb_duplicates() == 0
            else f" including {ms.sequences[0].nb_duplicates()} duplicates"
        )
        print(
            f"\n🗂️  All pictures belong to a single sequence ({len(ms.sequences[0].pictures)} pictures{d})"
        )

    return ms



import datetime
from typing import Optional
import re
from geovisio_cli import exception


def parse_hms(value: str, ref: int) -> Optional[float]:
    r"""Compute decimal coordinates (latitude or longitude) from degrees/minutes/seconds coordinates.

    Args:
        value: raw lat (or lon)
        ref: 1 if the value is North or East

    Returns:
        float: real lat (or lon) or None if a problem occurs
    >>> parse_hms("50° 30' 54.36\"", 1)
    50.5151
    >>> parse_hms("plop", 1) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    geovisio_cli.exception.CliException: impossible to parse position: plop
    """
    result = re.search(r"(\d+)° (\d+)' (.+?)\"", value)
    if not result:
        raise exception.CliException(f"impossible to parse position: {value}")
    h, m, s = result.groups()
    return (float(h) + float(m) / 60 + float(s) / 3600) * ref


def check_lat(value: Optional[str]) -> Optional[float]:
    """parse and check latitude"""
    if not value:
        return None
    try:
        f = float(value)
    except ValueError as e:
        raise exception.CliException(f"Impossible to parse latitude ({str(e)})")
    if f < -90 or f > 90:
        raise exception.CliException(
            f"latitude '{f}' is out of WGS84 bounds (should be in [-90, 90]"
        )
    return f


def check_lon(value: Optional[str]) -> Optional[float]:
    """parse and check longitude"""
    if not value:
        return None
    try:
        f = float(value)
    except ValueError as e:
        raise exception.CliException(f"Impossible to parse longitude ({str(e)})")
    if f < -180 or f > 180:
        raise exception.CliException(
            f"longitude '{f}' is out of WGS84 bounds (should be in [-180, 180]"
        )
    return f


def parse_capture_time(value: Optional[str]) -> Optional[float]:
    """
    Parse an RFC 3339 formated date and return the timestamp
    """
    if not value:
        return None
    try:
        d = datetime.datetime.fromisoformat(value)
    except Exception as e:
        raise exception.CliException(
            f"The capture_time was not recognized (should follow the RFC 3339): {value} ({str(e)})"
        )
    return d.timestamp()



import pytest
import os
from geovisio_cli.sequences.external_metadata import csv_metadata
from geovisio_cli import exception
from .conftest import FIXTURE_DIR
from geopic_tag_reader.reader import PartialGeoPicTags
from pathlib import Path


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "metadata.csv"),
)
def test_CsvMetadataHandler(datafiles):
    mtd = csv_metadata.CsvMetadataHandler(Path(datafiles / "metadata.csv"))
    assert mtd.data == {
        "e1.jpg": PartialGeoPicTags(lat=50.5151, lon=3.265, ts=1516589529.0),
        "e3.jpg": PartialGeoPicTags(lat=50.513433333, lon=3.265277778, ts=1516589649.0),
    }


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "metadata_exif.csv"),
)
def test_CsvMetadataHandler_exif(datafiles):
    mtd = csv_metadata.CsvMetadataHandler(Path(datafiles / "metadata_exif.csv"))
    assert mtd.data == {
        "e1.jpg": PartialGeoPicTags(
            lat=50.5151,
            lon=3.265,
            ts=1516589529.0,
            exif={
                "Exif.Image.Software": "MS Paint",
                "Exif.Image.Artist": "A 2 years old",
                "Xmp.xmp.Rating": "1",
            },
        ),
        "e3.jpg": PartialGeoPicTags(lat=50.513433333, lon=3.265277778, ts=1516589649.0),
    }


def test_bad_lat_CsvMetadataHandler(datafiles):
    p = datafiles / "bad_csv.csv"
    with open(p, "w") as f:
        f.write(
            r"""file,lon,lat,capture_time,field_of_view,heading
e1.jpg,3.265,lat,2018-01-22T02:52:09+00:00,,
"""
        )
    with pytest.raises(exception.CliException) as e:
        csv_metadata.CsvMetadataHandler(Path(p))
    assert (
        str(e.value)
        == "Impossible to parse latitude (could not convert string to float: 'lat')"
    )


def test_empty_file(datafiles):
    p = datafiles / "bad_csv.csv"
    with open(p, "w") as f:
        f.write(r"")
    with pytest.raises(exception.CliException) as e:
        csv_metadata.CsvMetadataHandler(Path(p))
    assert str(e.value) == "Invalid csv file: (Could not determine delimiter)"


def test_missing_column(datafiles):
    p = datafiles / "bad_csv.csv"
    with open(p, "w") as f:
        f.write(
            r"""plop,pouet
12,14
"""
        )
    with pytest.raises(exception.CliException) as e:
        csv_metadata.CsvMetadataHandler(Path(p))
    assert (
        str(e.value) == "Missing mandatory column 'file' to identify the picture's file"
    )


def test_bad_date_CsvMetadataHandler(datafiles):
    p = datafiles / "bad_csv.csv"
    with open(p, "w") as f:
        f.write(
            r"""file,lon,lat,capture_time,field_of_view,heading
e1.jpg,3.265,12,plop,,
"""
        )
    with pytest.raises(exception.CliException) as e:
        csv_metadata.CsvMetadataHandler(Path(p))
    assert (
        str(e.value)
        == "The capture_time was not recognized (should follow the RFC 3339): plop (Invalid isoformat string: 'plop')"
    )


@pytest.mark.parametrize(("delimiter"), ((","), (";"), ("\t")))
def test_tsv_delimiter(datafiles, delimiter):
    """Using a ';' should also be a valid delmiter"""
    p = datafiles / "bad_csv.csv"
    with open(p, "w") as f:
        r = r"""file;lon;lat;capture_time;field_of_view;heading
e1.jpg;3.265;50.5151;2018-01-22T02:52:09+00:00;;
e3.jpg;3.265277778;50.513433333;2018-01-22T02:54:09+00:00;360;15
""".replace(
            ";", delimiter
        )
        f.write(r)
    mtd = csv_metadata.CsvMetadataHandler(Path(p))
    assert mtd.data == {
        "e1.jpg": PartialGeoPicTags(lat=50.5151, lon=3.265, ts=1516589529.0),
        "e3.jpg": PartialGeoPicTags(lat=50.513433333, lon=3.265277778, ts=1516589649.0),
    }



# pyinstaller hook to find all pyexiv2 libraries
from PyInstaller.utils.hooks import collect_data_files, collect_dynamic_libs

# Collect the required binary files
binaries = collect_dynamic_libs("pyexiv2", search_patterns=["*.dll", "*.so", "*.pyd"])

# Collect any data files if needed
datas = collect_data_files("pyexiv2")



import os
import pytest

import geovisio_cli.sequences.status
import geovisio_cli.sequences.upload
from ..conftest import FIXTURE_DIR
from pathlib import Path
import requests
from geovisio_cli import exception, model
from datetime import timedelta
import tomli
from geopic_tag_reader.reader import PartialGeoPicTags
from datetime import datetime, timezone
import shutil


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_valid_upload(geovisio_with_token, datafiles):
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        pictureUploadTimeout=20,
    )
    assert len(uploadReports) == 1
    assert len(uploadReports[0].uploaded_pictures) == 3
    assert len(uploadReports[0].errors) == 0

    collection = model.Sequence(location=uploadReports[0].location)
    with requests.session() as s:
        geovisio_cli.sequences.status.wait_for_sequence(
            s, collection, timeout=timedelta(minutes=1)
        )
        status = geovisio_cli.sequences.status.get_status(s, collection)
        sequence_info = geovisio_cli.sequences.status.info(s, collection)
    # 3 pictures should have been uploaded
    assert len(status.pictures) == 3

    for i in status.pictures:
        assert i.status == "ready"

    # the collection should also have 3 items
    collection = requests.get(f"{collection.location}/items")
    collection.raise_for_status()

    features = collection.json()["features"]
    assert len(features) == 3

    assert sequence_info.title == "some title"
    assert sequence_info.interior_orientation == [
        model.InteriorOrientation(make="SONY", model="FDR-X1000V", field_of_view=None)
    ]

    # Check generated TOML file
    tomlFile = datafiles / model.SEQUENCE_TOML_FILE
    with open(tomlFile, "rb") as f:
        seqToml = tomli.load(f)
        assert seqToml["1"]["sequence"]["id"] == sequence_info.id
        assert seqToml["1"]["pictures"]["e1.jpg"]["id"] == features[0]["id"]
        assert seqToml["1"]["pictures"]["e2.jpg"]["id"] == features[1]["id"]
        assert seqToml["1"]["pictures"]["e3.jpg"]["id"] == features[2]["id"]


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_resume_upload(geovisio_with_token, datafiles):
    # Make e2 not valid to have a partial upload
    picE2 = datafiles / "e2.jpg"
    picE2bak = datafiles / "e2.bak"
    os.rename(picE2, picE2bak)
    with open(picE2, "w") as picE2file:
        picE2file.write("")
        picE2file.close()

    # Start upload -> 2 uploaded pics + 1 failure
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        sortMethod=model.SortMethod.filename_asc,
        pictureUploadTimeout=20,
    )
    assert len(uploadReports) == 1
    assert len(uploadReports[0].uploaded_pictures) == 2
    assert len(uploadReports[0].errors) == 1
    assert uploadReports[0].errors[0].position == 2

    # Check TOML file -> e2 has no ID but broken status
    tomlFile = datafiles / model.SEQUENCE_TOML_FILE
    with open(tomlFile, "rb") as f:
        seqToml = tomli.load(f)
        f.close()
        assert seqToml["1"]["pictures"]["e2.jpg"].get("id") is None
        assert seqToml["1"]["pictures"]["e2.jpg"]["status"] == "broken"

    # Make e2 valid
    os.remove(picE2)
    os.rename(picE2bak, picE2)

    # Launch again upload : 1 uploaded pic + 0 failure
    uploadReports2 = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        sortMethod=model.SortMethod.filename_asc,
        pictureUploadTimeout=20,
    )
    assert uploadReports2[0].location == uploadReports[0].location
    assert len(uploadReports2[0].uploaded_pictures) == 1
    assert len(uploadReports2[0].errors) == 0

    # Check TOML file -> everything has ID and looks like a charm
    with open(tomlFile, "rb") as f:
        seqToml2 = tomli.load(f)
        f.close()
        assert seqToml2["1"]["sequence"]["id"] == seqToml["1"]["sequence"]["id"]
        assert (
            seqToml2["1"]["pictures"]["e1.jpg"]["id"]
            == seqToml["1"]["pictures"]["e1.jpg"]["id"]
        )
        assert seqToml2["1"]["pictures"]["e2.jpg"]["id"] is not None
        assert seqToml2["1"]["pictures"]["e2.jpg"].get("status") is None
        assert (
            seqToml2["1"]["pictures"]["e3.jpg"]["id"]
            == seqToml["1"]["pictures"]["e3.jpg"]["id"]
        )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_upload_twice(geovisio_with_token, datafiles):
    """Uploading twice the same sequence, should result of nothing done in the second upload"""
    # Start upload -> 2 uploaded pics + 1 failure
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        pictureUploadTimeout=20,
    )
    assert len(uploadReports) == 1
    assert len(uploadReports[0].uploaded_pictures) == 3
    assert len(uploadReports[0].errors) == 0

    # Launch again upload : 1 uploaded pic + 0 failure
    uploadReports2 = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        pictureUploadTimeout=20,
    )
    assert len(uploadReports2) == 1
    assert uploadReports2[0].location == uploadReports[0].location
    assert len(uploadReports2[0].uploaded_pictures) == 0
    assert len(uploadReports2[0].errors) == 0


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_upload_with_duplicates(geovisio_with_token, datafiles):
    # Create a duplicate of e1
    shutil.copyfile(datafiles / "e1.jpg", datafiles / "e0.jpg")

    # Start upload
    uploadReports = geovisio_cli.sequences.upload.upload(
        path=Path(datafiles),
        geovisio=geovisio_with_token,
        title="some title",
        pictureUploadTimeout=20,
        mergeParams=geovisio_cli.sequences.process.standard.MergeParams(
            maxDistance=1, maxRotationAngle=30
        ),
    )
    assert len(uploadReports) == 1
    assert len(uploadReports[0].skipped_pictures) == 1
    assert len(uploadReports[0].uploaded_pictures) == 3
    assert len(uploadReports[0].errors) == 0


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_host(datafiles):
    with pytest.raises(exception.CliException) as e:
        geovisio_cli.sequences.upload.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(url="http://some_invalid_url"),
            title="some title",
            pictureUploadTimeout=20,
        )
    msg = str(e.value)
    assert msg.startswith(
        """The API is not reachable. Please check error and used URL below, and retry later if the URL is correct.

[bold]Used URL:[/bold] http://some_invalid_url/api
[bold]Error:[/bold]"""
    )

    # First one for local testing, second one for CI...
    assert (
        "Name or service not known" in msg
        or "Failed to resolve 'some_invalid_url'" in msg
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_path(geovisio_with_token, datafiles):
    with pytest.raises(exception.CliException) as e:
        geovisio_cli.sequences.upload.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(
                url=geovisio_with_token.url + "/some_additional_path"
            ),
            title=None,
            pictureUploadTimeout=20,
        )
    msg = str(e.value)
    assert msg.startswith(
        f"""The API URL is not valid.

Note that your URL should be the API root (something like https://geovisio.fr, https://panoramax.ign.fr or any other geovisio instance).
Please make sure you gave the correct URL and retry.

[bold]Used URL:[/bold] {geovisio_with_token.url}/some_additional_path/api
[bold]Error:[/bold]"""
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
)
def test_upload_on_invalid_url_schema(datafiles):
    with pytest.raises(exception.CliException) as e:
        geovisio_cli.sequences.upload.upload(
            path=Path(datafiles),
            geovisio=model.Geovisio(url="a non valid url at all"),
            title=None,
            pictureUploadTimeout=20,
        )
    assert str(e.value).startswith(
        """Error while connecting to the API. Please check error and used URL below

[bold]Used URL:[/bold] a non valid url at all/api
[bold]Error:[/bold]"""
    )


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "e1.jpg"),
    os.path.join(FIXTURE_DIR, "e2.jpg"),
    os.path.join(FIXTURE_DIR, "e3.jpg"),
)
def test_valid_upload_with_external_metadata(geovisio_with_token, datafiles):
    """
    Integration test on upload with external metada.
    """
    e1_new_dt = datetime(
        year=2023, month=9, day=1, hour=12, minute=45, second=36, tzinfo=timezone.utc
    )
    e3_new_dt = datetime(
        year=2023, month=9, day=1, hour=13, minute=45, second=36, tzinfo=timezone.utc
    )
    with requests.session() as s:
        assert geovisio_cli.sequences.upload.login(s, geovisio_with_token)
        sequences = model.ManySequences(
            datafiles / "_geovisio.toml",
            [
                model.Sequence(
                    title="some sequence",
                    pictures=[
                        model.Picture(
                            path=str(datafiles / "e1.jpg"),
                            overridden_metadata=PartialGeoPicTags(
                                lat=42, lon=12, ts=e1_new_dt.timestamp()
                            ),
                        ),
                        model.Picture(
                            path=str(datafiles / "e2.jpg"),
                            overridden_metadata=PartialGeoPicTags(
                                ts=e3_new_dt.timestamp()
                            ),
                        ),
                        model.Picture(
                            path=str(datafiles / "e3.jpg"),
                        ),
                    ],
                )
            ],
        )

        uploadReport = geovisio_cli.sequences.upload._publish(
            session=s,
            sequences=sequences,
            sequenceId=0,
            geovisio=geovisio_with_token,
            wait=False,
            alreadyBlurred=False,
            pictureUploadTimeout=2,
        )
        assert len(uploadReport.uploaded_pictures) == 3
        assert len(uploadReport.errors) == 0

        collection = model.Sequence(location=uploadReport.location)
        with requests.session() as s:
            geovisio_cli.sequences.status.wait_for_sequence(
                s, collection, timeout=timedelta(minutes=1)
            )
            status = geovisio_cli.sequences.status.get_status(s, collection)
            sequence_info = geovisio_cli.sequences.status.info(s, collection)
        # 3 pictures should have been uploaded
        assert len(status.pictures) == 3

        for i in status.pictures:
            assert i.status == "ready"

        # the collection should also have 3 items
        collection = requests.get(f"{collection.location}/items")
        collection.raise_for_status()

        features = collection.json()["features"]
        assert len(features) == 3
        # and the send pictures should have the overriden info
        assert features[0]["geometry"]["coordinates"] == [12, 42]
        assert features[0]["properties"]["datetime"] == "2023-09-01T12:45:36+00:00"
        assert features[1]["properties"]["datetime"] == "2023-09-01T13:45:36+00:00"

        assert sequence_info.title == "some sequence"

        # Check generated TOML file
        tomlFile = datafiles / model.SEQUENCE_TOML_FILE
        with open(tomlFile, "rb") as f:
            seqToml = tomli.load(f)
            f.close()
            assert seqToml["1"]["sequence"]["id"] == sequence_info.id
            assert seqToml["1"]["pictures"]["e1.jpg"]["id"] == features[0]["id"]
            assert seqToml["1"]["pictures"]["e2.jpg"]["id"] == features[1]["id"]
            assert seqToml["1"]["pictures"]["e3.jpg"]["id"] == features[2]["id"]
            # toml file should also have overriden metadata info
            assert seqToml["1"]["pictures"]["e1.jpg"]["overriden_metadata"] == {
                "lat": 42,
                "lon": 12,
                "ts": e1_new_dt.timestamp(),
            }
            assert seqToml["1"]["pictures"]["e2.jpg"]["overriden_metadata"] == {
                "ts": e3_new_dt.timestamp(),
            }



from __future__ import annotations
from geovisio_cli.sequences import external_metadata
from geovisio_cli.sequences.external_metadata import utils
from geovisio_cli.exception import CliException
from pathlib import Path
import csv
from geopic_tag_reader import reader
from typing import Dict, Optional, Tuple


def check(reader: csv.DictReader):
    if not reader.fieldnames or "file" not in reader.fieldnames:
        raise CliException(
            "Missing mandatory column 'file' to identify the picture's file"
        )


class CsvMetadataHandler(external_metadata.MetadataHandler):
    def __init__(self, file_name: Path) -> None:
        super().__init__()
        self.data = self._parse_file(file_name)

    @staticmethod
    def new_from_file(file_name: Path) -> Optional[CsvMetadataHandler]:
        if file_name.suffix != ".csv":
            return None

        return CsvMetadataHandler(file_name)

    def _parse_file(self, file_name: Path) -> Dict[str, reader.PartialGeoPicTags]:
        data = {}
        with open(file_name, "r") as f:
            # use Sniffer to detect the dialect of the file (separator, ...)
            try:
                dialect = csv.Sniffer().sniff(f.read(1024))
            except Exception as e:
                raise CliException(f"Invalid csv file: ({e})")

            f.seek(0)
            reader = csv.DictReader(f, dialect=dialect)
            check(reader)

            for row in reader:
                val = self.row_to_tag(row)
                if val:
                    pic_name, tag = val
                    data[pic_name] = tag

        return data

    def row_to_tag(self, row) -> Optional[Tuple[str, reader.PartialGeoPicTags]]:
        pic_name = row["file"]

        tags = reader.PartialGeoPicTags()

        tags.lat = utils.check_lat(row.get("lat"))
        tags.lon = utils.check_lon(row.get("lon"))
        tags.ts = utils.parse_capture_time(row.get("capture_time"))

        # Look for Exif/XMP columns
        tags.exif = {}
        for k in row:
            if k.startswith("Exif.") or k.startswith("Xmp."):
                v = row[k]
                if v is not None and len(v.strip()) > 0:
                    tags.exif[k] = row[k]

        return pic_name, tags

    def get(self, file_path: Path) -> Optional[reader.PartialGeoPicTags]:
        file_name = str(file_path.name)

        return self.data.get(str(file_name))



from datetime import timedelta
from time import sleep
from typing import Optional
from rich import print
from rich.console import Group
from rich.live import Live
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Table
from geovisio_cli import utils
from geovisio_cli.exception import CliException
from geovisio_cli.model import InteriorOrientation, Picture, Sequence


import requests


def get_status(session: requests.Session, sequence: Sequence) -> Sequence:
    s = session.get(
        f"{sequence.location}/geovisio_status", timeout=utils.REQUESTS_TIMEOUT
    )
    if s.status_code == 404:
        raise CliException(f"Sequence {sequence.location} not found")
    if s.status_code >= 400:
        raise CliException(
            f"Impossible to get sequence {sequence.location} status: {s.text}"
        )
    r = s.json()

    if len(sequence.pictures) == 0:
        sequence.pictures = [
            Picture(id=p["id"], status=p["status"])
            for p in r["items"]
            if p.get("id") is not None
        ]
    else:
        for i, p in enumerate(r["items"]):
            sequence.pictures[i].id = p["id"]
            sequence.pictures[i].status = p["status"]

    return sequence


def info(session: requests.Session, sequence: Sequence) -> Sequence:
    if not sequence.location:
        raise CliException(f"Sequence has no location set")

    s = session.get(sequence.location, timeout=utils.REQUESTS_TIMEOUT, verify=False)
    if s.status_code == 404:
        raise CliException(f"Sequence {sequence.location} not found")
    if s.status_code >= 400:
        raise CliException(
            f"Impossible to get sequence {sequence.location} status: {s.text}"
        )
    r = s.json()
    producer = next(
        (p["name"] for p in r.get("providers", []) if "producer" in p["roles"]), None
    )
    summary = r.get("summaries", {}).get("pers:interior_orientation", [])

    sequence.id = r["id"]
    sequence.title = r["title"]
    sequence.producer = producer
    sequence.interior_orientation = [
        InteriorOrientation(
            make=s.get("make"),
            model=s.get("model"),
            field_of_view=s.get("field_of_view"),
        )
        for s in summary
    ]

    return sequence


def _print_final_sequence_status(sequence: Sequence):
    nb_broken = sequence.nb_broken()
    nb_ready = sequence.nb_ready()
    if nb_ready == 0:
        print(f"[red]💥 No picture processed")
        return
    s = f"✅ {nb_ready} pictures processed"
    if nb_broken:
        s += f"([red]{nb_broken}[/red] cannot be processed)"
    print(s)


def wait_for_sequence(
    session: requests.Session, sequence: Sequence, timeout: Optional[timedelta] = None
):
    seq_status = get_status(session, sequence)
    if seq_status.all_done():
        _print_final_sequence_status(seq_status)
        return

    print("🔭 Waiting for pictures to be processed by geovisio")
    status_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(),
        TimeElapsedColumn(),
        MofNCompleteColumn(),
        "•",
        TextColumn("{task.fields[processing]}"),
    )
    processing_task = status_progress.add_task(
        f"[green]⏳ Processing ...",
        total=1,
        processing="",
    )
    progress_group = Group(
        status_progress,
    )
    waiting_time = timedelta(seconds=2)
    elapsed = timedelta(0)

    with Live(progress_group):
        while True:
            # TODO: display some stats about those errors

            nb_preparing = seq_status.nb_preparing()
            nb_waiting = seq_status.nb_waiting()
            processing = f"{nb_preparing} picture{('s' if nb_preparing != 0 else '')} currently processed"
            status_progress.update(
                processing_task,
                total=len(seq_status.pictures),
                completed=seq_status.nb_ready(),
                processing=processing,
            )

            if nb_waiting + nb_preparing == 0:
                break

            elapsed += waiting_time
            if timeout is not None and elapsed > timeout:
                raise CliException(f"❌ Sequence not ready after {elapsed}, stoping")

            sleep(waiting_time.total_seconds())
            seq_status = get_status(session, sequence)

    _print_final_sequence_status(seq_status)


def display_sequence_status(session: requests.Session, sequence: Sequence):
    seq_status = get_status(session, sequence)
    seq_info = info(session, sequence)

    s = f"Sequence [bold]{seq_info.title}[/bold]"
    if seq_info.producer is not None:
        s += f" produced by [bold]{seq_info.producer}[/bold]"
    s += " taken with"
    for i in seq_info.interior_orientation:
        s += f" [bold]{i.make} {i.model}[/bold]"
        if i.field_of_view:
            s += f" ({i.field_of_view}°)"

    print(s)
    table = Table()

    table.add_column("Total")
    table.add_column("Ready", style="green")
    table.add_column("Waiting", style="magenta")
    table.add_column("Preparing", style="magenta")
    table.add_column("Broken", style="red")

    table.add_row(
        f"{len(seq_status.pictures)}",
        f"{seq_status.nb_ready()}",
        f"{seq_status.nb_waiting()}",
        f"{seq_status.nb_preparing()}",
        f"{seq_status.nb_broken()}",
    )
    print(table)



import geovisio_cli.model
import pytest
from geovisio_cli.sequences import status
from geovisio_cli import exception
import requests


def test_status_on_unknown_collection(geovisio):
    with pytest.raises(exception.CliException) as e:
        with requests.session() as s:
            status.get_status(
                s,
                geovisio_cli.model.Sequence(
                    location=f"{geovisio.url}/api/collections/some_bad_id"
                ),
            )
    assert e.match(f"Sequence {geovisio.url}/api/collections/some_bad_id not found")
